import os
from datetime import date, datetime
from typing import Union

import jinja2

# variables that should be always available in a template
BASIC_VARS = dict(
    env=os.environ,
    now=datetime.now,
    utcnow=datetime.utcnow,
    today=date.today,
)


def finalize_placeholder(x):
    # This is used to make the `path` arg available in the filters and actions.
    # If a template uses `path` where no syspath is available this makes it possible
    # to raise an exception.
    if isinstance(x, Exception):
        raise x
    return x


Template = jinja2.Environment(
    variable_start_string="{",
    variable_end_string="}",
    autoescape=False,
    finalize=finalize_placeholder,
    undefined=jinja2.StrictUndefined,
)


def render(template: Union[str, jinja2.Template], args=None) -> str:
    if args is None:
        args = dict()
    try:
        if isinstance(template, jinja2.Template):
            text = template.render(**args, **BASIC_VARS)
        else:
            text = Template.from_string(template).render(**args, **BASIC_VARS)
    except jinja2.UndefinedError as e:
        msg = f"Missing value for template: {e}. Maybe you forgot a filter?"
        raise ValueError(msg) from e

    # expand user and fill environment vars
    text = os.path.expanduser(text)
    text = os.path.expandvars(text)
    return text



import subprocess
import sys
from datetime import datetime, timezone
from pathlib import Path
from typing import ClassVar, Optional

from organize.filter import FilterConfig

from .common.timefilter import TimeFilter


def read_stat_created(path: Path) -> Optional[int]:
    commands = (
        ["stat", "--format=%W", str(path)],  # GNU coreutils
        ["stat", "-f %B", str(path)],  # BSD
    )
    for cmd in commands:
        try:
            created_str = subprocess.check_output(cmd, encoding="utf-8").strip()
            timestamp = int(created_str)
            return timestamp
        except subprocess.CalledProcessError:
            pass
    return None


def read_created(path: Path) -> datetime:
    timestamp = None
    stat_result = path.stat()

    # ctime is the creation time only in Windows.
    # On unix it's the datetime of the last metadata change.
    if sys.platform == "win32":
        timestamp = stat_result.st_ctime
    else:
        # On other Unix systems (such as FreeBSD), the following
        # attributes may be available (but may be only filled out if
        # root tries to use them):
        try:
            timestamp = stat_result.st_birthtime  # type: ignore
        except AttributeError:
            pass

    # If we still haven't gotten a timestamp, we try the (slower) fallback
    # method using the `stat` tool.
    if timestamp is None:
        timestamp = read_stat_created(path)

    # give up.
    if timestamp is None:
        raise EnvironmentError("The creation time is not available.")

    return datetime.fromtimestamp(timestamp, timezone.utc)


class Created(TimeFilter):
    """Matches files / folders by created date

    Attributes:
        years (int): specify number of years
        months (int): specify number of months
        weeks (float): specify number of weeks
        days (float): specify number of days
        hours (float): specify number of hours
        minutes (float): specify number of minutes
        seconds (float): specify number of seconds
        mode (str):
            either 'older' or 'newer'. 'older' matches files / folders created before
            the given time, 'newer' matches files / folders created within the given
            time. (default = 'older')

    Returns:
        `{created}` (datetime): The datetime the file / folder was created.
    """

    filter_config: ClassVar[FilterConfig] = FilterConfig(
        name="created",
        files=True,
        dirs=True,
    )

    def get_datetime(self, path: Path) -> datetime:
        return read_created(path)



from __future__ import annotations

import os
import textwrap
from pathlib import Path
from typing import Iterable, List, Optional, Union

import yaml
from pydantic import ConfigDict, ValidationError
from pydantic.dataclasses import dataclass

from .errors import ConfigError
from .output import Default, Output
from .rule import Rule
from .template import render
from .utils import ReportSummary, normalize_unicode

Tags = Iterable[str]


def default_yaml_cnst(loader, tag_suffix, node):
    # disable yaml constructors for strings starting with exclamation marks
    # https://stackoverflow.com/a/13281292/300783
    return str(node.tag)


yaml.add_multi_constructor("", default_yaml_cnst, Loader=yaml.SafeLoader)


def should_execute(rule_tags: Tags, tags: Tags, skip_tags: Tags) -> bool:
    """
    returns whether the rule with `rule_tags` should be executed,
    given `tags` and `skip_tags`
    """
    if not rule_tags:
        rule_tags = set()
    if not tags:
        tags = set()
    if not skip_tags:
        skip_tags = set()

    if "always" in rule_tags and "always" not in skip_tags:
        return True
    if "never" in rule_tags and "never" not in tags:
        return False
    if not tags and not skip_tags:
        return True
    if not rule_tags and tags:
        return False
    should_run = any(tag in tags for tag in rule_tags) or not tags or not rule_tags
    should_skip = any(tag in skip_tags for tag in rule_tags)
    return should_run and not should_skip


@dataclass(config=ConfigDict(extra="ignore"))
class Config:
    rules: List[Rule]

    _config_path: Optional[Path] = None

    @classmethod
    def from_string(cls, config: str, config_path: Optional[Path] = None) -> Config:
        normalized = normalize_unicode(config)
        dedented = textwrap.dedent(normalized)
        as_dict = yaml.load(dedented, Loader=yaml.SafeLoader)
        try:
            inst = cls(**as_dict)
            inst._config_path = config_path
            return inst
        except ValidationError as e:
            # add a config_path property to the ValidationError
            raise ConfigError(e=e, config_path=config_path) from e

    @classmethod
    def from_path(cls, config_path: Path) -> Config:
        text = config_path.read_text(encoding="utf-8")
        inst = cls.from_string(text, config_path=config_path)
        return inst

    def execute(
        self,
        simulate: bool = True,
        output: Output = Default(),
        tags: Tags = set(),
        skip_tags: Tags = set(),
        working_dir: Union[str, Path] = ".",
    ) -> None:
        working_path = Path(render(str(working_dir)))
        os.chdir(working_path)
        output.start(
            simulate=simulate,
            config_path=self._config_path,
            working_dir=working_path,
        )
        summary = ReportSummary()
        try:
            for rule_nr, rule in enumerate(self.rules):
                if should_execute(
                    rule_tags=rule.tags,
                    tags=tags,
                    skip_tags=skip_tags,
                ):
                    rule_summary = rule.execute(
                        simulate=simulate,
                        output=output,
                        rule_nr=rule_nr,
                    )
                    summary += rule_summary
        finally:
            output.end(summary.success, summary.errors)



import textwrap
from typing import ClassVar, Dict, Optional

from pydantic import field_validator
from pydantic.config import ConfigDict
from pydantic.dataclasses import dataclass

from organize.filter import FilterConfig
from organize.output import Output
from organize.resource import Resource


@dataclass(config=ConfigDict(coerce_numbers_to_str=True, extra="forbid"))
class Python:

    """Use python code to filter files.

    Attributes:
        code (str):
            The python code to execute. The code must contain a `return` statement.


    **Returns:**

    - If your code returns `False` or `None` the file is filtered out,
      otherwise the file is passed on to the next filters.
    - `{python}` contains the returned value. If you return a dictionary (for
      example `return {"some_key": some_value, "nested": {"k": 2}}`) it will be
      accessible via dot syntax actions: `{python.some_key}`, `{python.nested.k}`.
    - Variables of previous filters are available, but you have to use the normal python
      dictionary syntax `x = regex["my_group"]`.
    """

    code: str

    filter_config: ClassVar[FilterConfig] = FilterConfig(
        name="python",
        files=True,
        dirs=True,
    )

    @field_validator("code", mode="after")
    @classmethod
    def must_have_return_statement(cls, value):
        if "return" not in value:
            raise ValueError("No return statement found in your code!")
        return value

    def __post_init__(self):
        self.code = textwrap.dedent(self.code)

    def __usercode__(self, print, **kwargs) -> Optional[Dict]:
        raise NotImplementedError()

    def pipeline(self, res: Resource, output: Output) -> bool:
        def _output_msg(*values, sep: str = " ", end: str = ""):
            """
            the print function for the use code needs to print via the current output
            """
            output.msg(
                res=res,
                msg=f"{sep.join(str(x) for x in values)}{end}",
                sender="python",
            )

        # codegen the user function with arguments as available in the resource
        kwargs = ", ".join(res.dict().keys())
        func = f"def __userfunc(print, {kwargs}):\n"
        func += textwrap.indent(self.code, "    ")
        func += "\n\nself.__usercode__ = __userfunc"

        exec(func, globals().copy(), locals().copy())
        result = self.__usercode__(print=_output_msg, **res.dict())

        if isinstance(result, dict):
            res.deep_merge(key=self.filter_config.name, data=result)
        else:
            res.vars[self.filter_config.name] = result
        return result not in (False, None)



import os
from itertools import chain
from pathlib import Path
from typing import Iterator, Optional, Tuple

import platformdirs

from organize.utils import expandvars

from .errors import ConfigNotFound

ENV_ORGANIZE_CONFIG = os.environ.get("ORGANIZE_CONFIG")
USER_CONFIG_DIR = platformdirs.user_config_path(appname="organize")
XDG_CONFIG_DIR = expandvars(os.environ.get("XDG_CONFIG_HOME", "~/.config")) / "organize"


def find_default_config() -> Path:
    # if the `ORGANIZE_CONFIG` env variable is set we only check this specific location
    if ENV_ORGANIZE_CONFIG is not None:
        result = expandvars(ENV_ORGANIZE_CONFIG)
        if result.exists() and result.is_file():
            return result
        raise ConfigNotFound(str(result), init_path=result)

    # no ORGANIZE_CONFIG env variable given:
    # -> check the default config location
    result = USER_CONFIG_DIR / "config.yaml"
    if result.exists() and result.is_file():
        return result
    raise ConfigNotFound(str(result), init_path=result)


def find_config(name_or_path: Optional[str] = None) -> Path:
    if name_or_path is None:
        return find_default_config()

    # otherwise we try:
    # 0. The full path if applicable
    # 1.`$PWD`
    # 2. the platform specifig config dir
    # 3. `$XDG_CONFIG_HOME/organize`
    as_path = expandvars(name_or_path)
    if as_path.exists() and as_path.is_file():
        return as_path

    search_pathes: Tuple[Path, ...] = tuple()
    if not as_path.is_absolute():
        as_yml = Path(f"{as_path}.yml")
        as_yaml = Path(f"{as_path}.yaml")
        search_pathes = (
            as_path,
            as_yaml,
            as_yml,
            USER_CONFIG_DIR / as_path,
            USER_CONFIG_DIR / as_yaml,
            USER_CONFIG_DIR / as_yml,
            XDG_CONFIG_DIR / as_path,
            XDG_CONFIG_DIR / as_yaml,
            XDG_CONFIG_DIR / as_yml,
        )
        for path in search_pathes:
            if path.exists() and path.is_file():
                return path

    if str(as_path).lower().endswith((".yaml", ".yml")):
        init_path = as_path
    else:
        init_path = USER_CONFIG_DIR / as_yaml
    raise ConfigNotFound(
        config=name_or_path,
        search_pathes=search_pathes,
        init_path=init_path,
    )


def list_configs() -> Iterator[Path]:
    for loc in (USER_CONFIG_DIR, XDG_CONFIG_DIR):
        yield from chain(loc.glob("*.yml"), loc.glob("*.yaml"))



import sys
from typing import ClassVar

import simplematch as sm
from pydantic.config import ConfigDict
from pydantic.dataclasses import dataclass

from organize.action import ActionConfig
from organize.output import Output
from organize.resource import Resource
from organize.template import Template, render
from organize.validators import FlatList


@dataclass(config=ConfigDict(coerce_numbers_to_str=True, extra="forbid"))
class MacOSTags:

    """Add macOS tags.

    Attributes:
        *tags (str): A list of tags or a single tag.

    The color can be specified in brackets after the tag name, for example:

    ```yaml
    macos_tags: "Invoices (red)"
    ```

    Available colors are `none`, `gray`, `green`, `purple`, `blue`, `yellow`, `red` and
    `orange`.
    """

    tags: FlatList[str]

    action_config: ClassVar[ActionConfig] = ActionConfig(
        name="macos_tags",
        standalone=False,
        files=True,
        dirs=True,
    )

    def __post_init__(self):
        self._tags = [Template.from_string(tag) for tag in self.tags]
        if sys.platform != "darwin":
            raise EnvironmentError("The macos_tags action is only available on macOS")

    def pipeline(self, res: Resource, output: Output, simulate: bool):
        import macos_tags

        COLORS = [c.name.lower() for c in macos_tags.Color]

        for template in self._tags:
            tag = render(template, res.dict())
            name, color = self._parse_tag(tag)

            if color not in COLORS:
                raise ValueError(
                    "color %s is unknown. (Available: %s)" % (color, " / ".join(COLORS))
                )

            output.msg(
                res=res,
                sender=self,
                msg=f'Adding tag: "{name}" (color: {color})',
            )
            if not simulate:
                _tag = macos_tags.Tag(
                    name=name,
                    color=macos_tags.Color[color.upper()],
                )  # type: ignore
                macos_tags.add(_tag, file=str(res.path))

    def _parse_tag(self, s):
        """parse a tag definition and return a tuple (name, color)"""
        result = sm.match("{name} ({color})", s)
        if not result:
            return s, "none"
        return result["name"], result["color"].lower()



import os
from fnmatch import fnmatch
from pathlib import Path
from typing import Iterable, Iterator, List, Literal, NamedTuple, Optional, Set

from pydantic import Field
from pydantic.dataclasses import dataclass


def pattern_match(name: str, patterns: Iterable[str]) -> bool:
    return any(fnmatch(name, pat) for pat in patterns)


class ScandirResult(NamedTuple):
    dirs: List[os.DirEntry]
    nondirs: List[os.DirEntry]


def scandir(top: str, collectfiles: bool = True) -> ScandirResult:
    result = ScandirResult(dirs=[], nondirs=[])
    try:
        # build iterator if we have the permissions to this folder
        scandir_it = os.scandir(top)
    except OSError:
        return result

    with scandir_it:
        while True:
            try:
                try:
                    entry = next(scandir_it)
                except StopIteration:
                    break
            except OSError:
                return result

            try:
                is_symlink = entry.is_symlink()
            except OSError:
                # If is_symlink() raises an OSError, consider that the
                # entry is not a symbolic link, same behaviour than
                # os.path.islink().
                is_symlink = False

            # As of now, we skip all symlinks.
            if is_symlink:
                continue

            try:
                is_dir = entry.is_dir()
            except OSError:
                # If is_dir() raises an OSError, consider that the entry is not
                # a directory, same behaviour than os.path.isdir().
                is_dir = False

            if is_dir:
                result.dirs.append(entry)
            elif collectfiles:
                result.nondirs.append(entry)
    return result


class DirActions(NamedTuple):
    to_yield: List[os.DirEntry]
    to_walk: List[os.DirEntry]


@dataclass(frozen=True)
class Walker:
    min_depth: int = 0
    max_depth: Optional[int] = None
    method: Literal["breadth", "depth"] = "breadth"
    filter_dirs: Optional[List[str]] = None
    filter_files: Optional[List[str]] = None
    exclude_dirs: Set[str] = Field(default_factory=set)
    exclude_files: Set[str] = Field(default_factory=set)

    def _should_yield_file(self, entry: os.DirEntry, lvl: int) -> bool:
        return (
            lvl >= self.min_depth
            and not pattern_match(entry.name, self.exclude_files)
            and (
                self.filter_files is None
                or pattern_match(entry.name, self.filter_files)
            )
        )

    def _dir_actions(self, entries: Iterable[os.DirEntry], lvl: int) -> DirActions:
        result = DirActions(to_yield=[], to_walk=[])
        for entry in entries:
            if not pattern_match(entry.name, self.exclude_dirs) and (
                self.filter_dirs is None or pattern_match(entry.name, self.filter_dirs)
            ):
                if self.max_depth is None or lvl < self.max_depth:
                    result.to_walk.append(entry)
                if lvl >= self.min_depth:
                    result.to_yield.append(entry)
        return result

    def walk(
        self,
        top: str,
        files: bool = True,
        dirs: bool = True,
        lvl: int = 0,
    ) -> Iterator[os.DirEntry]:
        if not files and not dirs:
            return

        # list all dirs and nondirs of the folder
        result = scandir(top, collectfiles=files)

        if self.method == "breadth":
            # Return entries
            for entry in result.nondirs:
                if files and self._should_yield_file(entry=entry, lvl=lvl):
                    yield entry
            dir_actions = self._dir_actions(result.dirs, lvl=lvl)
            if dirs:
                yield from dir_actions.to_yield
            # Recurse into sub-directories
            for entry in dir_actions.to_walk:
                yield from self.walk(entry.path, files=files, dirs=dirs, lvl=lvl + 1)

        elif self.method == "depth":
            dir_actions = self._dir_actions(result.dirs, lvl=lvl)
            # Recurse into sub-directories
            for entry in dir_actions.to_walk:
                yield from self.walk(entry.path, files=files, dirs=dirs, lvl=lvl + 1)
            # Return entries
            for entry in result.nondirs:
                if files and self._should_yield_file(entry=entry, lvl=lvl):
                    yield entry
            if dirs:
                yield from dir_actions.to_yield
        else:
            raise ValueError(f'Unknown method "{self.method}"')

    def files(self, path: str) -> Iterator[Path]:
        # if path is a single file we emit just the path itself
        if os.path.isfile(path):
            yield Path(path)
            return
        # otherwise we walk the given folder
        for entry in self.walk(path, files=True, dirs=False):
            yield Path(entry.path)

    def dirs(self, path: str) -> Iterator[Path]:
        for entry in self.walk(path, files=False, dirs=True):
            yield Path(entry.path)



from typing import Any, ClassVar, List, Union

import simplematch
from pydantic.config import ConfigDict
from pydantic.dataclasses import dataclass

from organize.filter import FilterConfig
from organize.output import Output
from organize.resource import Resource
from organize.utils import normalize_unicode


@dataclass(config=ConfigDict(coerce_numbers_to_str=True, extra="forbid"))
class Name:
    """Match files and folders by name

    Attributes:
        match (str):
            A matching string in [simplematch-syntax](https://github.com/tfeldmann/simplematch)

        startswith (str):
            The filename must begin with the given string

        contains (str):
            The filename must contain the given string

        endswith (str):
            The filename (without extension) must end with the given string

        case_sensitive (bool):
            By default, the matching is case sensitive. Change this to False to use
            case insensitive matching.
    """

    match: str = "*"
    startswith: Union[str, List[str]] = ""
    contains: Union[str, List[str]] = ""
    endswith: Union[str, List[str]] = ""
    case_sensitive: bool = True

    filter_config: ClassVar[FilterConfig] = FilterConfig(
        name="name",
        files=True,
        dirs=True,
    )

    def __post_init__(self, *args, **kwargs):
        self._matcher = simplematch.Matcher(
            self.match,
            case_sensitive=self.case_sensitive,
        )
        self.startswith = self.create_list(self.startswith, self.case_sensitive)
        self.contains = self.create_list(self.contains, self.case_sensitive)
        self.endswith = self.create_list(self.endswith, self.case_sensitive)

    def matches(self, name: str) -> bool:
        if not self.case_sensitive:
            name = name.lower()

        is_match = (
            self._matcher.test(name)
            and any(x in name for x in self.contains)
            and any(name.startswith(x) for x in self.startswith)
            and any(name.endswith(x) for x in self.endswith)
        )
        return is_match

    def pipeline(self, res: Resource, output: Output) -> bool:
        assert res.path is not None, "Does not support standalone mode"
        if res.is_dir():
            name = res.path.stem
        else:
            name, ext = res.path.stem, res.path.suffix
            if not name:
                name = ext
        result = self.matches(normalize_unicode(name))
        m = self._matcher.match(normalize_unicode(name))
        if not m:
            m = name

        res.vars[self.filter_config.name] = m
        return result

    @staticmethod
    def create_list(x: Union[int, str, List[Any]], case_sensitive: bool) -> List[str]:
        if isinstance(x, (int, float)):
            x = str(x)
        if isinstance(x, str):
            x = [x]
        x = [str(x) for x in x]
        if not case_sensitive:
            x = [x.lower() for x in x]
        return x



"""
Duplicate detection filter.

Based on this stackoverflow answer:
    https://stackoverflow.com/a/36113168/300783

Which was updated for python3 in:
    https://gist.github.com/tfeldmann/fc875e6630d11f2256e746f67a09c1ae
"""
from collections import defaultdict
from pathlib import Path
from typing import Any, Callable, ClassVar, Literal, NamedTuple, Tuple

from pydantic.config import ConfigDict
from pydantic.dataclasses import dataclass

from organize.filter import FilterConfig
from organize.filters.created import read_created
from organize.filters.hash import hash, hash_first_chunk
from organize.filters.lastmodified import read_lastmodified
from organize.filters.size import read_file_size
from organize.output import Output
from organize.resource import Resource

DetectionMethod = Literal[
    "first_seen",
    "-first_seen",
    "last_seen",
    "-last_seen",
    "name",
    "-name",
    "created",
    "-created",
    "lastmodified",
    "-lastmodified",
]


class OriginalDetectionResult(NamedTuple):
    original: Path
    duplicate: Path

    @classmethod
    def by_sorting(
        cls,
        known: Path,
        new: Path,
        key: Callable[[Path], Any],
    ) -> "OriginalDetectionResult":
        if key(known) <= key(new):
            return cls(original=known, duplicate=new)
        return cls(original=new, duplicate=known)

    def reversed(self) -> "OriginalDetectionResult":
        return OriginalDetectionResult(
            original=self.duplicate,
            duplicate=self.original,
        )


def detect_original(
    known: Path, new: Path, method: DetectionMethod, reverse: bool
) -> Tuple[Path, Path]:
    """Returns a tuple (original file, duplicate)"""

    if method == "first_seen":
        result = OriginalDetectionResult(original=known, duplicate=new)
    elif method == "last_seen":
        result = OriginalDetectionResult(original=new, duplicate=known)
    elif method == "name":
        result = OriginalDetectionResult.by_sorting(
            known=known, new=new, key=lambda x: x.name
        )
    elif method == "created":
        result = OriginalDetectionResult.by_sorting(
            known=known, new=new, key=lambda x: read_created(x)
        )
    elif method == "lastmodified":
        result = OriginalDetectionResult.by_sorting(
            known=known, new=new, key=lambda x: read_lastmodified(x)
        )
    else:
        raise ValueError(f"Unknown original detection method: {method}")

    return result.reversed() if reverse else result


@dataclass(config=ConfigDict(extra="forbid"))
class Duplicate:
    """A fast duplicate file finder.

    This filter compares files byte by byte and finds identical files with potentially
    different filenames.

    Attributes:
        detect_original_by (str):
            Detection method to distinguish between original and duplicate.
            Possible values are:

            - `"first_seen"`: Whatever file is visited first is the original. This
              depends on the order of your location entries.
            - `"name"`: The first entry sorted by name is the original.
            - `"created"`: The first entry sorted by creation date is the original.
            - `"lastmodified"`: The first file sorted by date of last modification is
               the original.

    You can reverse the sorting method by prefixing a `-`.

    So with `detect_original_by: "-created"` the file with the older creation date is
    the original and the younger file is the duplicate. This works on all methods, for
    example `"-first_seen"`, `"-name"`, `"-created"`, `"-lastmodified"`.

    **Returns:**

    `{duplicate.original}` - The path to the original
    """

    detect_original_by: DetectionMethod = "first_seen"
    hash_algorithm: str = "sha1"

    filter_config: ClassVar[FilterConfig] = FilterConfig(
        name="duplicate", files=True, dirs=False
    )

    def __post_init__(self):
        # reverse original detection order if starting with "-"
        self._detect_original_by = self.detect_original_by
        self._detect_original_reverse = False
        if self.detect_original_by.startswith("-"):
            self._detect_original_by = self.detect_original_by[1:]
            self._detect_original_reverse = True

        self._files_for_size = defaultdict(list)
        self._files_for_chunk = defaultdict(list)
        self._file_for_hash = dict()

        # we keep track of the files we already computed the hashes for so we only do
        # that once.
        self._seen_files = set()
        self._first_chunk_known = set()
        self._hash_known = set()

    def pipeline(self, res: Resource, output: Output) -> bool:
        assert res.path is not None, "Does not support standalone mode"
        # skip symlinks
        if res.path.is_symlink():
            return False

        # the exact same path has already been handled. This happens if multiple
        # locations emit this file in a single rule or if we follow symlinks.
        # We skip these.
        if res.path in self._seen_files:
            return False

        self._seen_files.add(res.path)

        # check for files with equal size
        file_size = read_file_size(path=res.path)
        same_size = self._files_for_size[file_size]
        same_size.append(res.path)
        if len(same_size) == 1:
            # the file is unique in size and cannot be a duplicate
            return False

        # for all other files with the same file size:
        # make sure we know their hash of their first 1024 byte chunk
        for f in same_size[:-1]:
            if f not in self._first_chunk_known:
                chunk_hash = hash_first_chunk(f, algo=self.hash_algorithm)
                self._first_chunk_known.add(f)
                self._files_for_chunk[chunk_hash].append(f)

        # check first chunk hash collisions with the current file
        chunk_hash = hash_first_chunk(res.path, algo=self.hash_algorithm)
        same_first_chunk = self._files_for_chunk[chunk_hash]
        same_first_chunk.append(res.path)
        self._first_chunk_known.add(res.path)
        if len(same_first_chunk) == 1:
            # the file has a unique small hash and cannot be a duplicate
            return False

        # Ensure we know the full hashes of all files with the same first chunk as
        # the investigated file
        for f in same_first_chunk[:-1]:
            if f not in self._hash_known:
                hash_ = hash(f, algo=self.hash_algorithm)
                self._hash_known.add(f)
                self._file_for_hash[hash_] = f

        # check full hash collisions with the current file
        hash_ = hash(res.path, algo=self.hash_algorithm)
        self._hash_known.add(res.path)
        known = self._file_for_hash.get(hash_)
        if known:
            original, duplicate = detect_original(
                known=known,
                new=res.path,
                method=self._detect_original_by,
                reverse=self._detect_original_reverse,
            )
            if known != original:
                self._file_for_hash[hash_] = original

            res.path = duplicate
            res.vars[self.filter_config.name] = {"original": original}
            return True

        return False



from typing import List

from .jsonl import JSONL, EventType


class SavingOutput(JSONL):
    """
    Saves all the incoming event messages of the latest run.
    """

    def __init__(self) -> None:
        self.queue: List[EventType] = []

    def start(self, *args, **kwargs) -> None:
        self.queue.clear()
        super().start(*args, **kwargs)

    def emit_event(self, event: EventType) -> None:
        self.queue.append(event)

    def _messages_of_kind(self, kind: str) -> List:
        return [x for x in self.queue if x.type == kind]

    @property
    def msg_start(self):
        result = self._messages_of_kind("START")
        if len(result) != 1:
            raise ValueError("Multiple start events found")
        return result[0]

    @property
    def msg_msg(self):
        return self._messages_of_kind("MSG")

    @property
    def msg_report(self):
        result = self._messages_of_kind("REPORT")
        if len(result) != 1:
            raise ValueError("Multiple reports found")
        return result[0]

    @property
    def messages(self):
        return [x.msg for x in self.queue if x.type == "MSG"]



import sys
from typing import ClassVar, List

from pydantic import Field, field_validator
from pydantic.config import ConfigDict
from pydantic.dataclasses import dataclass

from organize.filter import FilterConfig
from organize.output import Output
from organize.resource import Resource
from organize.utils import glob_match


def list_tags(path) -> List[str]:
    import macos_tags

    tags = macos_tags.get_all(path)
    return ["{} ({})".format(tag.name, tag.color.name.lower()) for tag in tags]


def matches_tags(filter_tags, file_tags) -> bool:
    if not filter_tags:
        return True
    if not file_tags:
        return False
    for tag in file_tags:
        if any(glob_match(filter_tag, tag) for filter_tag in filter_tags):
            return True
    return False


@dataclass(config=ConfigDict(coerce_numbers_to_str=True, extra="forbid"))
class MacOSTags:
    """Filter by macOS tags

    Attributes:
        tags (list(str) or str):
            The tags to filter by
    """

    tags: List[str] = Field(default_factory=list)

    filter_config: ClassVar[FilterConfig] = FilterConfig(
        name="macos_tags",
        files=True,
        dirs=True,
    )

    def __post_init__(self):
        if sys.platform != "darwin":
            raise EnvironmentError("The macos_tags filter is only available on macOS")

    @field_validator("tags", mode="before")
    def ensure_list(cls, v):
        if isinstance(v, str):
            return [v]
        return v

    def pipeline(self, res: Resource, output: Output) -> bool:
        file_tags = list_tags(res.path)
        res.vars[self.filter_config.name] = file_tags
        return matches_tags(filter_tags=self.tags, file_tags=file_tags)



import logging
import re
from pathlib import Path
from typing import Any, ClassVar

from pydantic.config import ConfigDict
from pydantic.dataclasses import dataclass

from organize.filter import FilterConfig
from organize.output import Output
from organize.resource import Resource


@dataclass(config=ConfigDict(coerce_numbers_to_str=True, extra="forbid"))
class FileContent:
    """Matches file content with the given regular expression

    Attributes:
        expr (str): The regular expression to be matched.

    Any named groups (`(?P<groupname>.*)`) in your regular expression will
    be returned like this:

    **Returns:**

    - `{filecontent.groupname}`: The text matched with the named group
      `(?P<groupname>)`
    """

    expr: str = r"(?P<all>.*)"

    filter_config: ClassVar[FilterConfig] = FilterConfig(
        name="filecontent",
        files=True,
        dirs=False,
    )

    def __post_init__(self):
        self._expr = re.compile(self.expr, re.MULTILINE | re.DOTALL)

    def matches(self, path: Path) -> Any:
        try:
            import textract

            content = textract.process(str(path), errors="ignore")
            match = self._expr.search(content.decode("utf-8", errors="ignore"))
            return match
        except ImportError as e:
            raise ImportError(
                "textract is not installed. "
                "Install with pip install organize-tool[textract]"
            ) from e
        except textract.exceptions.CommandLineError as e:
            logging.exception(e)

    def pipeline(self, res: Resource, output: Output) -> bool:
        assert res.path is not None, "Does not support standalone mode"
        match = self.matches(path=res.path)

        if match:
            res.deep_merge(self.filter_config.name, match.groupdict())
        return bool(match)



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Dict, Optional

from rich.console import Console
from rich.panel import Panel
from rich.prompt import Confirm as RichConfirm
from rich.status import Status
from rich.theme import Theme

from organize.utils import ChangeDetector, escape

from ._sender import sender_name
from .output import Level

if TYPE_CHECKING:
    from typing import List

    from organize.resource import Resource

    from ._sender import SenderType


def format_path(path: Path, base_style: str, main_style: str) -> str:
    base = escape(f"{path.parent}/")
    main = escape(path.name)
    return f"[{base_style}]{base}[/][{main_style}]{main}[/]"


MSG_STYLE: Dict[Level, str] = {
    "info": "[pipeline.msg]",
    "warn": "[pipeline.warn]",
    "error": "[pipeline.error]ERROR! ",
}


def format_msg(
    msg: str,
    level: Level,
    sender: SenderType,
    standalone: bool,
    escape_msg: bool = True,
) -> str:
    _msg = escape(msg) if escape_msg else msg
    indent = "    - " if not standalone else ""

    fmt_sender = f"([pipeline.source]{escape(sender_name(sender))}[/])"
    fmt_msg = f"{MSG_STYLE[level]}{_msg}[/]"
    return f"{indent}{fmt_sender} {fmt_msg}"


class Confirm(RichConfirm):
    @classmethod
    def set_error_msg(cls, msg: str) -> None:
        cls.validate_error_message = msg


class Default:
    def __init__(self, theme: Optional[Theme] = None, errors_only: bool = False):
        if theme is None:
            theme = Theme(
                {
                    "info": "dim cyan",
                    "warning": "yellow",
                    "error": "bold red",
                    "simulation": "bold green",
                    "status": "bold green",
                    "rule": "bold cyan",
                    "location.base": "green",
                    "location.main": "bold green",
                    "path.base": "dim green",
                    "path.main": "green",
                    "pipeline.source": "cyan",
                    "pipeline.msg": "",
                    "pipeline.warn": "yellow",
                    "pipeline.error": "bold red",
                    "pipeline.prompt": "bold yellow",
                    "summary.done": "bold green",
                    "summary.fail": "red",
                }
            )
        self.errors_only = errors_only
        self.msg_queue: List[str] = []
        self.det_resource = ChangeDetector()

        self.console = Console(theme=theme, highlight=False)

        self.status = Status("", console=self.console)
        self.det_rule = ChangeDetector()
        self.det_location = ChangeDetector()
        self.det_path = ChangeDetector()
        self.simulate = False

    def show_resource(self, res: Resource):
        # rule changed
        if self.det_rule.changed(res.rule):
            self.det_location.reset()
            self.det_path.reset()
            self.console.print()
            rule_name = f"Rule #{res.rule_nr}"
            if res.rule is not None and res.rule.name is not None:
                rule_name += f": {res.rule.name}"
            self.console.rule(
                f"[rule]:gear: {escape(rule_name)}",
                align="left",
                style="rule",
            )

        # location changed
        if self.det_location.changed(res.basedir):
            self.det_path.reset()
            if res.basedir:
                path_str = format_path(
                    Path(res.basedir),
                    "location.base",
                    "location.main",
                )
                self.console.print(path_str)

        # path changed
        if self.det_path.changed(res.path):
            relative_path = res.relative_path()
            if relative_path is not None:
                path_str = format_path(relative_path, "path.base", "path.main")
                self.console.print(f"  {path_str}")

    def start(
        self,
        simulate: bool,
        config_path: Optional[Path],
        working_dir: Path,
    ) -> None:
        self.det_rule.reset()
        self.det_location.reset()
        self.det_path.reset()

        self.simulate = simulate
        if self.simulate:
            self.console.print(Panel("SIMULATION", style="simulation"))

        if working_dir.resolve() != Path(".").resolve():
            self.console.print(f'Working dir: "{escape(working_dir)}"')
        if config_path:
            self.console.print(f'Config: "{escape(config_path)}"')

        status_verb = "simulating" if simulate else "organizing"
        self.status.update(f"[status]{status_verb}[/]")
        self.status.start()

    def msg(
        self,
        res: Resource,
        msg: str,
        sender: SenderType,
        level: Level = "info",
    ) -> None:
        msg = format_msg(
            msg=msg,
            level=level,
            sender=sender,
            standalone=res.path is None,
        )

        if self.errors_only:
            if self.det_resource.changed(res):
                self.msg_queue.clear()
            self.msg_queue.append(msg)
            if level == "error":
                self.show_resource(res)
                for msg in self.msg_queue:
                    self.console.print(msg)
                self.msg_queue.clear()
        else:
            self.show_resource(res)
            self.console.print(msg)

    def confirm(
        self,
        res: Resource,
        msg: str,
        default: bool,
        sender: SenderType,
    ) -> bool:
        self.status.stop()
        self.show_resource(res)
        msg_prompt = format_msg(
            msg=f"[pipeline.prompt]{escape(msg)}[/]",
            level="info",
            sender=sender,
            standalone=res.path is None,
            escape_msg=False,
        )
        msg_invalid = format_msg(
            msg="[prompt.invalid]Please enter Y or N[/]",
            level="info",
            sender=sender,
            standalone=res.path is None,
            escape_msg=False,
        )
        Confirm.set_error_msg(msg=msg_invalid)
        result = Confirm.ask(prompt=msg_prompt, console=self.console, default=default)
        self.status.start()
        return result

    def end(self, success_count: int, error_count: int):
        self.status.stop()
        self.console.print()
        if success_count == 0 and error_count == 0:
            self.console.print("[summary.done]Nothing to do[/]")
        else:
            msg = (
                f"[summary.done]success {success_count}[/] / "
                f"[summary.fail]fail {error_count}[/]"
            )
            self.console.print(msg)
        if self.simulate:
            self.console.print(Panel("SIMULATION", style="simulation"))



from datetime import datetime, tzinfo
from pathlib import Path
from typing import ClassVar, Literal, Union

import arrow
from pydantic.config import ConfigDict
from pydantic.dataclasses import dataclass

from organize.filter import FilterConfig
from organize.output import Output
from organize.resource import Resource


@dataclass(config=ConfigDict(extra="forbid", arbitrary_types_allowed=True))
class TimeFilter:
    years: int = 0
    months: int = 0
    weeks: int = 0
    days: int = 0
    hours: int = 0
    minutes: int = 0
    seconds: int = 0
    mode: Literal["older", "newer"] = "older"
    timezone: Union[tzinfo, str] = "local"

    filter_config: ClassVar[FilterConfig] = FilterConfig(
        "timefilter",
        files=True,
        dirs=False,
    )

    def __post_init__(self):
        self._has_comparison = (
            self.years
            or self.months
            or self.weeks
            or self.days
            or self.hours
            or self.minutes
            or self.seconds
        )
        self._comparison_dt = (
            arrow.now()
            .shift(
                years=-self.years,
                months=-self.months,
                weeks=-self.weeks,
                days=-self.days,
                hours=-self.hours,
                minutes=-self.minutes,
                seconds=-self.seconds,
            )
            .datetime
        )

    def matches_datetime(self, dt: datetime) -> bool:
        if not self._has_comparison:
            return True

        if self.mode == "older":
            return dt < self._comparison_dt
        elif self.mode == "newer":
            return dt > self._comparison_dt
        else:
            raise ValueError(f"Unknown mode {self.mode}")

    def pipeline(self, res: Resource, output: Output) -> bool:
        assert res.path is not None, "Does not support standalone mode"
        try:
            dt = self.get_datetime(res.path)
        except Exception:
            return False

        # apply timezone
        dt = arrow.get(dt).to(self.timezone).datetime

        res.vars[self.filter_config.name] = dt
        return self.matches_datetime(dt)

    def get_datetime(self, path: Path) -> datetime:
        raise NotImplementedError()



from typing import ClassVar

from pydantic.config import ConfigDict
from pydantic.dataclasses import dataclass

from organize.action import ActionConfig
from organize.output import Output
from organize.resource import Resource
from organize.template import Template, render

from .common.conflict import ConflictMode, resolve_conflict
from .common.target_path import prepare_target_path


@dataclass(config=ConfigDict(coerce_numbers_to_str=True, extra="forbid"))
class Symlink:

    """Create a symbolic link.

    Attributes:
        dest (str):
            The symlink destination. If **dest** ends with a slash `/``, create the
            symlink in the given directory. Can contain placeholders.

        on_conflict (str):
            What should happen in case **dest** already exists.
            One of `skip`, `overwrite`, `trash`, `rename_new` and `rename_existing`.
            Defaults to `rename_new`.

        rename_template (str):
            A template for renaming the file / dir in case of a conflict.
            Defaults to `{name} {counter}{extension}`.

        autodetect_folder (bool):
            In case you forget the ending slash "/" to indicate creating the
            link inside the destination folder this settings will handle targets
            without a file extension as folders.  If you really mean to copy to
            a file without file extension, set this to false.
            Default: true
    """

    dest: str
    on_conflict: ConflictMode = "rename_new"
    rename_template: str = "{name} {counter}{extension}"
    autodetect_folder: bool = True

    action_config: ClassVar[ActionConfig] = ActionConfig(
        name="symlink",
        standalone=False,
        files=True,
        dirs=True,
    )

    def __post_init__(self):
        self._dest = Template.from_string(self.dest)
        self._rename_template = Template.from_string(self.rename_template)

    def pipeline(self, res: Resource, output: Output, simulate: bool):
        assert res.path is not None, "Does not support standalone mode"
        rendered = render(self._dest, res.dict())
        dst = prepare_target_path(
            src_name=res.path.name,
            dst=rendered,
            autodetect_folder=self.autodetect_folder,
            simulate=simulate,
        )

        skip_action, dst = resolve_conflict(
            dst=dst,
            res=res,
            conflict_mode=self.on_conflict,
            rename_template=self._rename_template,
            simulate=simulate,
            output=output,
        )
        if skip_action:
            return

        output.msg(res=res, msg=f"Creating symlink at {dst}", sender=self)
        res.walker_skip_pathes.add(dst)
        if not simulate:
            dst.symlink_to(target=res.path, target_is_directory=res.is_dir())



from pathlib import Path
from typing import ClassVar

from pydantic.config import ConfigDict
from pydantic.dataclasses import dataclass

from organize.action import ActionConfig
from organize.output import Output
from organize.resource import Resource


def trash(path: Path):
    from send2trash import send2trash

    send2trash(path)


@dataclass(config=ConfigDict(coerce_numbers_to_str=True, extra="forbid"))
class Trash:

    """Move a file or dir into the trash."""

    action_config: ClassVar[ActionConfig] = ActionConfig(
        name="trash",
        standalone=False,
        files=True,
        dirs=True,
    )

    def pipeline(self, res: Resource, output: Output, simulate: bool):
        assert res.path is not None, "Does not support standalone mode"
        output.msg(res=res, msg=f'Trash "{res.path}"', sender=self)
        if not simulate:
            trash(res.path)



from pathlib import Path


def user_wants_a_folder(path: str, autodetect: bool) -> bool:
    """
    Try to detect whether the user means to target a folder
    """
    if path.endswith(("/", "\\")):
        return True

    if autodetect:
        return "." not in Path(path).name

    return False


def prepare_target_path(
    src_name: str,
    dst: str,
    autodetect_folder: bool,
    simulate: bool,
) -> Path:
    result = Path(dst)
    wants_folder = user_wants_a_folder(path=dst, autodetect=autodetect_folder)

    # if dst is an existing folder, we use it
    if result.exists():
        if result.is_dir():
            return result / src_name
        elif wants_folder:
            raise ValueError(f'Expected "{dst}" to be a folder, but it\'s not!')

    if wants_folder:
        if not simulate:
            result.mkdir(parents=True, exist_ok=True)
        return result / src_name
    else:
        if not simulate:
            result.parent.mkdir(parents=True, exist_ok=True)
        return result



from pathlib import Path
from typing import ClassVar, Set, Tuple

from pydantic import Field, field_validator
from pydantic.config import ConfigDict
from pydantic.dataclasses import dataclass

from organize.filter import FilterConfig
from organize.output import Output
from organize.resource import Resource
from organize.validators import flatten


def convert_to_list(v):
    if not v:
        return []
    if isinstance(v, str):
        return v.split()
    return v


def normalize_extension(ext: str) -> str:
    """strip colon and convert to lowercase"""
    if ext.startswith("."):
        return ext[1:].lower()
    else:
        return ext.lower()


@dataclass(config=ConfigDict(coerce_numbers_to_str=True, extra="forbid"))
class Extension:
    """Filter by file extension

    Attributes:
        *extensions (list(str) or str):
            The file extensions to match (does not need to start with a colon).

    **Returns:**

    - `{extension}`: the original file extension (without colon)
    """

    extensions: Set[str] = Field(default_factory=set)

    filter_config: ClassVar[FilterConfig] = FilterConfig(
        name="extension",
        files=True,
        dirs=False,
    )

    @field_validator("extensions", mode="before")
    def normalize_extensions(cls, v):
        as_list = convert_to_list(v)
        return set(map(normalize_extension, flatten(list(as_list))))

    def suffix_match(self, path: Path) -> Tuple[str, bool]:
        suffix = path.suffix.lstrip(".")
        if not self.extensions:
            return (suffix, True)
        if not suffix:
            return (suffix, False)
        return (suffix, normalize_extension(suffix) in self.extensions)

    def pipeline(self, res: Resource, output: Output) -> bool:
        assert res.path is not None, "Does not support standalone mode"
        if res.is_dir():
            raise ValueError("Dirs not supported")

        suffix, match = self.suffix_match(path=res.path)
        res.vars[self.filter_config.name] = suffix
        return match



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, ClassVar, Literal

from pydantic.config import ConfigDict
from pydantic.dataclasses import dataclass

from organize.action import ActionConfig
from organize.template import Template, render

if TYPE_CHECKING:
    from organize.output import Output
    from organize.resource import Resource


@dataclass(config=ConfigDict(coerce_numbers_to_str=True, extra="forbid"))
class Write:

    """
    Write text to a file.

    If the specified path does not exist it will be created.

    Attributes:
        text (str):
            The text that should be written. Supports templates.

        outfile (str):
            The file `text` should be written into. Supports templates.

        mode (str):
            Can be either `append` (append text to the file), `prepend` (insert text as
            first line) or `overwrite` (overwrite content with text).
            Defaults to `append`.

        encoding (str):
            The text encoding to use. Default: "utf-8".

        newline (str):
            (Optional) Whether to append a newline to the given `text`.
            Defaults to `true`.

        clear_before_first_write (bool):
            (Optional) Clears the file before first appending / prepending text to it.
            This happens only the first time the file is written to. If the rule filters
            don't match anything the file is left as it is.
            Defaults to `false`.
    """

    text: str
    outfile: str
    mode: Literal["append", "prepend", "overwrite"] = "append"
    encoding: str = "utf-8"
    newline: bool = True
    clear_before_first_write: bool = False

    action_config: ClassVar[ActionConfig] = ActionConfig(
        name="write",
        standalone=True,
        files=True,
        dirs=True,
    )

    def __post_init__(self):
        self._text = Template.from_string(self.text)
        self._path = Template.from_string(self.outfile)
        self._known_files = set()

    def pipeline(self, res: Resource, output: Output, simulate: bool):
        text = render(self._text, res.dict())
        path = Path(render(self._path, res.dict()))

        resolved = path.resolve()
        if resolved not in self._known_files:
            self._known_files.add(resolved)

            if not simulate:
                resolved.parent.mkdir(parents=True, exist_ok=True)

            # clear on first write
            if resolved.exists() and self.clear_before_first_write:
                output.msg(res=res, msg=f"Clearing file {path}", sender=self)
                if not simulate:
                    resolved.open("w")  # clear the file

        output.msg(res=res, msg=f'{path}: {self.mode} "{text}"', sender=self)
        if self.newline:
            text += "\n"

        if not simulate:
            if self.mode == "append":
                with open(path, "a", encoding=self.encoding) as f:
                    f.write(text)
            elif self.mode == "prepend":
                content = ""
                if path.exists():
                    content = path.read_text(encoding=self.encoding)
                path.write_text(text + content, encoding=self.encoding)
            elif self.mode == "overwrite":
                path.write_text(text, encoding=self.encoding)



from pathlib import Path
from typing import Dict, Iterable, List, Literal, Optional, Set

from pydantic import BaseModel, ConfigDict, Field, field_validator, model_validator

from .action import Action
from .filter import All, Any, Filter, HasFilterPipeline, Not
from .location import Location
from .output import Output
from .registry import action_by_name, filter_by_name
from .resource import Resource
from .template import render
from .utils import ReportSummary
from .validators import FlatList, flatten
from .walker import Walker

FilterMode = Literal["all", "any", "none"]


def action_from_dict(d: Dict) -> Action:
    """
    :param d:
        A dict in the forms of
        { "action_name": None }
        { "action_name": "value" }
        { "action_name": {"param": "value"} }
    :returns:
        An instantiated action.
    """
    if not len(d.keys()) == 1:
        raise ValueError("Action definition must have only one key")
    name, value = next(iter(d.items()))
    ActionCls = action_by_name(name)
    if value is None:
        return ActionCls()
    elif isinstance(value, dict):
        return ActionCls(**value)
    else:
        return ActionCls(value)


def filter_from_dict(d: Dict) -> Filter:
    """
    :param d:
        A dict in the forms of ("not" prefix is optional)
        { "[not] filter_name": None }
        { "[not] filter_name": "value" }
        { "[not] filter_name": {"param": "value"} }
    :returns: An instantiated filter.
    """
    if not len(d.keys()) == 1:
        raise ValueError("Filter definition must have a single key")
    name, value = next(iter(d.items()))

    # check for "not" in filter key
    invert_filter = False
    if name.startswith("not "):
        name = name[4:]
        invert_filter = True

    FilterCls = filter_by_name(name)

    # instantiate
    if value is None:
        inst = FilterCls()
    elif isinstance(value, dict):
        inst = FilterCls(**value)
    else:
        inst = FilterCls(value)

    return Not(inst) if invert_filter else inst


def filter_pipeline(
    filters: Iterable[Filter],
    filter_mode: FilterMode,
    res: Resource,
    output: Output,
) -> bool:
    collection: HasFilterPipeline
    if filter_mode == "all":
        collection = All(*filters)
    elif filter_mode == "any":
        collection = Any(*filters)
    elif filter_mode == "none":
        collection = All(*[Not(x) for x in filters])
    else:
        raise ValueError(f"Unknown filter mode {filter_mode}")
    return collection.pipeline(res, output=output)


def action_pipeline(
    actions: Iterable[Action],
    res: Resource,
    simulate: bool,
    output: Output,
) -> Iterable[Action]:
    for action in actions:
        try:
            yield action
            action.pipeline(res=res, simulate=simulate, output=output)
        except StopIteration:
            break


class Rule(BaseModel):
    name: Optional[str] = None
    enabled: bool = True
    targets: Literal["files", "dirs"] = "files"
    locations: FlatList[Location] = Field(default_factory=list)
    subfolders: bool = False
    tags: Set[str] = Field(default_factory=set)
    filters: List[Filter] = Field(default_factory=list)
    filter_mode: FilterMode = "all"
    actions: List[Action] = Field(..., min_length=1)

    model_config = ConfigDict(
        extra="forbid",
        arbitrary_types_allowed=True,
    )

    @field_validator("locations", mode="before")
    def validate_locations(cls, locations):
        if locations is None:
            return []
        locations = flatten(locations)
        result = []
        for x in locations:
            if isinstance(x, str):
                x = {"path": x}
            result.append(x)
        return result

    @field_validator("filters", mode="before")
    def validate_filters(cls, filters):
        result = []
        filters = flatten(filters)
        for x in filters:
            # make sure "- extension" becomes "- extension:"
            if isinstance(x, str):
                x = {x: None}
            # create instance from dict
            if isinstance(x, dict):
                result.append(filter_from_dict(x))
            # other instances
            else:
                result.append(x)
        return result

    @field_validator("actions", mode="before")
    def validate_actions(cls, actions):
        result = []
        actions = flatten(actions)
        for x in actions:
            # make sure "- extension" becomes "- extension:"
            if isinstance(x, str):
                x = {x: None}
            # create instance from dict
            if isinstance(x, dict):
                result.append(action_from_dict(x))
            # other instances
            else:
                result.append(x)
        return result

    @model_validator(mode="after")
    def validate_target_support(self) -> "Rule":
        # standalone mode
        if not self.locations:
            if self.filters:
                raise ValueError("Filters are present but no locations are given!")
            for action in self.actions:
                if not action.action_config.standalone:
                    raise ValueError(
                        f'Action "{action.action_config.name}" does not support '
                        "standalone mode (no rule.locations specified)."
                    )
        # targets dirs
        if self.targets == "dirs":
            for filter in self.filters:
                if not filter.filter_config.dirs:
                    raise ValueError(
                        f'Filter "{filter.filter_config.name}" does not support '
                        "folders (targets: dirs)"
                    )
            for action in self.actions:
                if not action.action_config.dirs:
                    raise ValueError(
                        f'Action "{action.action_config.name}" does not support '
                        "folders (targets: dirs)"
                    )
        # targets files
        elif self.targets == "files":
            for filter in self.filters:
                if not filter.filter_config.files:
                    raise ValueError(
                        f'Filter "{filter.filter_config.name}" does not support '
                        "files (targets: files)"
                    )
            for action in self.actions:
                if not action.action_config.files:
                    raise ValueError(
                        f'Action "{action.action_config.name}" does not support '
                        "files (targets: files)"
                    )
        else:
            raise ValueError(f"Unknown target: {self.targets}")

        return self

    def walk(self, rule_nr: int = 0):
        for location in self.locations:
            # instantiate the filesystem walker
            exclude_files = location.system_exclude_files | location.exclude_files
            exclude_dirs = location.system_exclude_dirs | location.exclude_dirs
            if location.max_depth == "inherit":
                max_depth = None if self.subfolders else 0
            else:
                max_depth = location.max_depth

            walker = Walker(
                min_depth=location.min_depth,
                max_depth=max_depth,
                filter_dirs=location.filter_dirs,
                filter_files=location.filter,
                method="breadth",
                exclude_dirs=exclude_dirs,
                exclude_files=exclude_files,
            )

            # whether to walk dirs or files
            _walk_funcs = {
                "files": walker.files,
                "dirs": walker.dirs,
            }
            for loc_path in location.path:
                expanded_path = render(loc_path)
                for path in _walk_funcs[self.targets](expanded_path):
                    yield Resource(
                        path=Path(path),
                        basedir=Path(expanded_path),
                        rule=self,
                        rule_nr=rule_nr,
                    )

    def execute(
        self, *, simulate: bool, output: Output, rule_nr: int = 0
    ) -> ReportSummary:
        if not self.enabled:
            return ReportSummary()

        # standalone mode
        if not self.locations:
            res = Resource(path=None, rule_nr=rule_nr)
            try:
                for action in action_pipeline(
                    actions=self.actions,
                    res=res,
                    simulate=simulate,
                    output=output,
                ):
                    pass
                return ReportSummary(success=1)
            except Exception as e:
                output.msg(
                    res=res,
                    msg=str(e),
                    level="error",
                    sender=action,
                )
                # logging.exception(e)
                return ReportSummary(errors=1)

        # normal mode
        summary = ReportSummary()
        skip_pathes: Set[Path] = set()
        for res in self.walk(rule_nr=rule_nr):
            if res.path in skip_pathes:
                continue
            result = filter_pipeline(
                filters=self.filters,
                filter_mode=self.filter_mode,
                res=res,
                output=output,
            )
            if result:
                try:
                    for action in action_pipeline(
                        actions=self.actions,
                        res=res,
                        simulate=simulate,
                        output=output,
                    ):
                        pass
                    skip_pathes = skip_pathes.union(res.walker_skip_pathes)
                    summary.success += 1
                except Exception as e:
                    output.msg(
                        res=res,
                        msg=str(e),
                        level="error",
                        sender=action,
                    )
                    # logging.exception(e)
                    summary.errors += 1
        return summary
